/*
 * Copyright (c) 2015-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>
#include <cstdio>
#include <cmath>
#include <iostream>

#include <Eigen/Geometry>

#include "pom_c_types.h"

#include "codels.h"
#include "ukf.h"


/* --- Task filter ------------------------------------------------------ */

/** Codel pom_filter_start of task filter.
 *
 * Triggered by pom_start.
 * Yields to pom_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_start(pom_context_s **context, const pom_frame *frame,
                 genom_context self)
{
  frame->open("robot", self);
  return pom_exec;
}


/** Codel pom_filter_exec of task filter.
 *
 * Triggered by pom_exec.
 * Yields to pom_pause_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_exec(pom_context_s **context, const pom_log_s *log,
                const pom_frame *frame, genom_context self)
{
  if (!(*context)) return pom_pause_exec;

  pom_history_s &h = (*context)->history;
  pom_history_s::reverse_iterator i;
  double date, mdate;

  /* get the oldest non fused state */
  for(i = h.rbegin(); i != h.rend(); i++)
    if (!i->second.fused) break;

  /* set to the fused state before */
  if (i == h.rbegin()) {
    (*context)->filter.set((*context)->state, (*context)->cov);
    date = (*context)->ts.sec + 1e-9 * (*context)->ts.nsec;
  } else {
    i--;
    (*context)->filter.set(i->second.state, i->second.cov);
    date = i->first.sec + 1e-9 * i->first.nsec;
    i++;
  }

  for(; i != h.rend(); i++) {
    pom_snapshot &s = i->second;

    mdate = s.measurement.ts.sec + 1e-9 * s.measurement.ts.nsec;
    if (mdate <= date) {
      (*context)->filter.predict(0.);
      (*context)->filter.update(s.state, s.cov);
      s.fused = true;
      continue;
    }

    (*context)->filter.predict(mdate - date);

    if (s.measurement.pos._present) {
      (*context)->filter.measure.p() <<
        s.measurement.pos._value.x,
        s.measurement.pos._value.y,
        s.measurement.pos._value.z;
      (*context)->filter.measure.q() <<
        s.measurement.pos._value.qx,
        s.measurement.pos._value.qy,
        s.measurement.pos._value.qz,
        s.measurement.pos._value.qw;

      if (s.measurement.pos_cov._present) {
        double *c = s.measurement.pos_cov._value.cov;

        (*context)->filter.measure.p_cov() <<
          c[0],  c[1],  c[3],
          c[1],  c[2],  c[4],
          c[3],  c[4],  c[5];
        (*context)->filter.measure.q_cov() <<
          c[14], c[19], c[25], c[13],
          c[19], c[20], c[26], c[18],
          c[25], c[26], c[27], c[24],
          c[13], c[18], c[24],  c[9];
      } else {
        /* no covariance provided, provide a pessimistic default */
        double qx = s.measurement.pos._value.qx;
        double qy = s.measurement.pos._value.qy;
        double qz = s.measurement.pos._value.qz;
        double qw = s.measurement.pos._value.qw;

        (*context)->filter.measure.p_cov() <<
          1.,  0.,  0.,
          0.,  1.,  0.,
          0.,  0.,  1.;
        (*context)->filter.measure.q_cov() <<
          0.25*(1-qx*qx), -0.25*qx*qy,     -0.25*qx*qz,     -0.25*qw*qx,
          -0.25*qx*qy,     0.25*(1-qy*qy), -0.25*qy*qz,     -0.25*qw*qy,
          -0.25*qx*qz,    -0.25*qy*qz,      0.25*(1-qz*qz), -0.25*qw*qz,
          -0.25*qw*qx,    -0.25*qw*qy,     -0.25*qw*qz,      0.25*(1-qw*qw);
      }
    }

    /* linear velocity */
    if (s.measurement.vel._present &&
        !std::isnan(s.measurement.vel._value.vx)) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iv() :
       (*context)->filter.measure.v()) <<
        s.measurement.vel._value.vx,
        s.measurement.vel._value.vy,
        s.measurement.vel._value.vz;

      if (s.measurement.vel_cov._present) {
        double *c = s.measurement.vel_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    /* angular velocity */
    if (s.measurement.vel._present &&
        !std::isnan(s.measurement.vel._value.wx)) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iw() :
       (*context)->filter.measure.w()) <<
        s.measurement.vel._value.wx,
        s.measurement.vel._value.wy,
        s.measurement.vel._value.wz;

      if (s.measurement.vel_cov._present) {
        double *c = s.measurement.vel_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          c[9],  c[13], c[18],
          c[13], c[14], c[19],
          c[18], c[19], c[20];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          0.25,  0.,    0.,
          0.,    0.25,  0.,
          0.,    0.,    0.25;
      }
    }

    /* linear acceleration - XXX when intrisic, assume accelerometer */
    if (s.measurement.acc._present) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iawg() :
       (*context)->filter.measure.a()) <<
        s.measurement.acc._value.ax,
        s.measurement.acc._value.ay,
        s.measurement.acc._value.az;

      if (s.measurement.acc_cov._present) {
        double *c = s.measurement.acc_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    (*context)->filter.update(s.state, s.cov);
    s.fused = true;

    (*context)->filter.set(s.state, s.cov);
    date = mdate;
  }

  /* current state */
  struct timeval now;
  gettimeofday(&now, NULL);

  (*context)->filter.predict(now.tv_sec + 1e-6*now.tv_usec - date);
  (*context)->filter.update((*context)->state, (*context)->cov);
  (*context)->ts.sec = now.tv_sec;
  (*context)->ts.nsec = now.tv_usec * 1000;


  /* publish */
  or_pose_estimator_state *r = frame->data("robot", self);
  if (r) {
    r->ts = (*context)->ts;

    r->pos._present = true;
    r->pos._value.x = (*context)->state.p()(0);
    r->pos._value.y = (*context)->state.p()(1);
    r->pos._value.z = (*context)->state.p()(2);
    r->pos._value.qw = (*context)->state.q()(3);
    r->pos._value.qx = (*context)->state.q()(0);
    r->pos._value.qy = (*context)->state.q()(1);
    r->pos._value.qz = (*context)->state.q()(2);
    {
      Eigen::Matrix<double, 7, 7> c((*context)->state.pq_cov((*context)->cov));
      r->pos_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 7; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->pos_cov._value.cov[k] = c(m, n);
    }

    r->vel._present = true;
    r->vel._value.vx = (*context)->state.v()(0);
    r->vel._value.vy = (*context)->state.v()(1);
    r->vel._value.vz = (*context)->state.v()(2);
    r->vel._value.wx = (*context)->state.w()(0);
    r->vel._value.wy = (*context)->state.w()(1);
    r->vel._value.wz = (*context)->state.w()(2);
    {
      Eigen::Matrix<double, 6, 6> c((*context)->state.vw_cov((*context)->cov));
      r->vel_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 6; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->vel_cov._value.cov[k] = c(m, n);
    }

    r->acc._present = true;
    r->acc._value.ax = (*context)->state.a()(0);
    r->acc._value.ay = (*context)->state.a()(1);
    r->acc._value.az = (*context)->state.a()(2);
    {
      Eigen::Matrix<double, 3, 3> c((*context)->state.a_cov((*context)->cov));
      r->acc_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 3; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->acc_cov._value.cov[k] = c(m, n);
    }
  }
  frame->write("robot", self);

  /* log */
  if (log->statef) {
    double qw = r->pos._value.qw;
    double qx = r->pos._value.qx, qy = r->pos._value.qy, qz = r->pos._value.qz;
    double roll, pitch, yaw;

    roll  = atan2(2 * (qw*qx + qy*qz), 1 - 2 * (qx*qx + qy*qy));
    pitch = asin(2 * (qw*qy - qz*qx));
    yaw   = atan2(2 * (qw*qz + qx*qy), 1 - 2 * (qy*qy + qz*qz));

    fprintf(
      log->statef, pom_log_state_fmt "\n",
      r->ts.sec, r->ts.nsec,
      r->intrinsic,
      r->pos._present, r->vel._present, r->acc._present,
      r->pos._value.x, r->pos._value.y, r->pos._value.z,
      roll, pitch, yaw,
      r->vel._value.vx, r->vel._value.vy, r->vel._value.vz,
      r->vel._value.wx, r->vel._value.wy, r->vel._value.wz,
      r->acc._value.ax, r->acc._value.ay, r->acc._value.az);
  }

  return pom_pause_exec;
}
